class VariantCounter(object):
    def __init__(self):
        self.counters = {}

    def observe(self, s):
        try:
            self.counters[s] += 1
        except KeyError:
            self.counters[s] = 1

    def results(self, matric):
        if matric == 'values':
            return self.counters.values()
        elif matric == 'keys':
            return self.counters.keys()
        else:
            return self.counters



import re
from ..dynamic_constants import DynamicConstants


class HeaderDecode(object):
    @staticmethod
    def _get_header_regex(header_format: str):
        componants = re.findall(r"\(.[^)]*\)", header_format)

        expression = header_format

        for idx, match in enumerate(componants, 1):
            param = re.match(r"\((.[^)]*)\)", match).group(1)
            expression = expression.replace(match, "(?P<{param}>.[^|]*){close}".format(param=param, close="\\" if idx != len(componants) else ""))

        return expression

    @staticmethod
    def set_header_regex(header_format: str):
        DynamicConstants.HEADER_REGEX = HeaderDecode._get_header_regex(header_format)

        return DynamicConstants.HEADER_REGEX



from collections import Counter, defaultdict


class VariantDict(defaultdict):
    """
        Data structure for storing sequence indexes

        Describes the data structure for storing the indexes of sequences where a specific variant originates from

        Methods:
            get_counter: Returns a Counter object for all the variants of the current position
    """

    def get_counter(self) -> Counter:
        """
            Get a counter object for the current position's variants

            :return: Returns a counter object for variants at current position
        """

        position = Counter()
        for x in self:
            position[x] = len(self[x])
        return position



class Variant(dict):
    def __init__(self, position, sequence, count, incidence, motif_short=None, motif_long=None):
        """
            Data Structure for kmer variants

            Describes the data structure for kmer variants

            :param position: The position kmer position within the sequence
            :param sequence: The sequence of the given variant
            :param count: The number of times the variant is seen in the alignment at this position
            :param incidence: The percentage conservation of this variant for this position
            :param motif_short: The short name motif classification for this variant
            :param motif_long: The long name motif classification for this variant

            :type position: int
            :type sequence: str
            :type count: int
            :type incidence: float
            :type motif_short: str
            :type motif_long: str
        """

        self.position = position
        self.sequence = sequence
        self.count = count
        self.incidence = incidence
        self.motif_short = motif_short
        self.motif_long = motif_long

    def __setattr__(self, key, value):
        dict.__setitem__(self, key, value)

    def __getattr__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            raise AttributeError

    def __getstate__(self):
        state = self.copy()
        state = {key: value for key, value in state.items() if value}

        return state



class SequenceFileNotFound(Exception):
    def __init__(self, seq_path: str):
        """
            Error raised when the FASTA file path is invalid, or the file is not found.
            
            :param seq_path: The path to the FASTA file.
            :type seq_path: str 
        """

        msg = f'\nThe FASTA file was not found at the following location:\n\t{seq_path}\nAdvice: \n\tPlease check ' \
              f'whether the path is valid, or contains any errors.'

        super(SequenceFileNotFound, self).__init__(msg)


class NoHeaderFormat(Exception):
    def __init__(self):
        """
            Error raised when header decoding is enabled, but no header format is provided.
        """

        msg = 'No header format provided. Please provide the header format.\n\tUsing command-line: Use the flag ' \
              '"hunana -f", or run "hunana -h" for help.\n\tUsing module: Set the "header_format" attribute.'

        super(NoHeaderFormat, self).__init__(msg)


class HeaderDecodeError(Exception):
    def __init__(self, header_format: str):
        """
            Error raised when the correct regex pattern could not be generated.

            :param header_format: The format of the FASTA header.
            :type header_format: str
        """

        msg = f'\nFailed to generate correct regex expression using the header format:\n\t{header_format}\nAdvice:\n\t' \
              f'Make sure the header format you provided accurately reflects the format of all the FASTA headers in ' \
              f'your FASTA file.'

        super(HeaderDecodeError, self).__init__(msg)


class SequenceLengthError(Exception):
    def __init__(self, groups_dict: dict):
        """
            Error raised when not all sequences in the slignment are of equal length.

            :param groups_dict: A dictionary containing sequences grouped by sequence length.
            :type groups_dict: dict
        """

        msg = '\nThe FASTA sequences provided does not appear to be aligned and are of various lengths.\n'

        error_data = []

        for length, sequences in groups_dict.items():
            headers = [sequence.description for sequence in sequences]
            headers_string = '\n\t'.join(headers)

            error_data.append(
                '\nLength: {length}\n\t{seqs}'.format(length=length, seqs=headers_string)
            )

        # Yes, this kind of string concatenation is bad, but it's a single time thing and doesn't cost much
        super(SequenceLengthError, self).__init__(msg + '\n'.join(error_data))


class NoSequencesProvided(Exception):
    def __init__(self):
        """
            Error raised when no sequences are provided by the user.
        """

        msg = 'No sequences are provided. Please make sure your FASTA file is not empty, or invalid.'

        super(NoSequencesProvided, self).__init__(msg)


class InvalidKmerLength(Exception):
    def __init__(self, seq_len: int, kmer_len: int):
        """
            Error raised when the kmer length is bigger than, or equal to the sequence length.

            :param seq_len: The length of the sequences provided by the user.
            :param kmer_len: The length of the kmers that user wants to generate.

            :type seq_len: int
            :type kmer_len: int
        """

        msg = f'The kmer length is bigger than, or equal to the length of the sequenced provided.\n\tKMER LENGTH: ' \
              f'{kmer_len}\n\tSEQUENCE LENGTH: {seq_len}'

        super(InvalidKmerLength, self).__init__(msg)


class HeaderItemCountInvalid(Exception):
    def __init__(self, header: str, desired_count: int, actual_count: int):
        """
            Error raised when the number of items provided in the header format does not match the number of items in
            the actual FASTA header of a sequence.

            :param header: The affected FASTA header.
            :param desired_count: The desired number of header items derived from the header format provided by user.
            :param actual_count: The actual number of header items in the FASTA header.

            :type header: str
            :type desired_count: int
            :type actual_count: int
        """

        msg = f'The header data count does not match.\n\tDESIRED HEADER DATA COUNT: {desired_count}\n\tACTUAL COUNT: ' \
              f'{actual_count}\n\tAFFECTED HEADER: {header}'

        super(HeaderItemCountInvalid, self).__init__(msg)


class HeaderItemEmpty(Exception):
    def __init__(self, header: str):
        """
            Error raised when header data errors are not suppressed (via no_header_data_error) and a header item
            is empty.

            :param header: The affected FASTA header.
            :type header: str
        """

        msg = f'Empty header items found:\n\tAFFECTED HEADER: {header}\n\nYou may set the flag ' \
              f'"no_header_error=True" to ignore this error and Hunana shall append "Unknown" to all empty ' \
              f'header items.'

        super(HeaderItemEmpty, self).__init__(msg)
